import pytest

from bookstore.client.nb_client import (
    NotebookSession,
    KernelInfo,
    extract_kernel_id,
    LiveNotebookRecord,
)

from bookstore.tests.client.client_fixtures import *


def test_notebook_server_record(notebook_server_record, notebook_server_dict):
    assert notebook_server_record.base_url == notebook_server_dict['base_url']
    assert notebook_server_record.hostname == notebook_server_dict["hostname"]
    assert notebook_server_record.notebook_dir == notebook_server_dict["notebook_dir"]
    assert notebook_server_record.password == notebook_server_dict['password']
    assert notebook_server_record.pid == notebook_server_dict["pid"]
    assert notebook_server_record.port == notebook_server_dict["port"]
    assert notebook_server_record.secure == notebook_server_dict["secure"]
    assert notebook_server_record.token == notebook_server_dict["token"]
    assert notebook_server_record.url == notebook_server_dict["url"]
    assert notebook_server_record == LiveNotebookRecord(**notebook_server_dict)


def test_kernel_info_class(kernel_info_dict, kernel_info):
    assert kernel_info.id == kernel_info_dict['id']
    assert kernel_info.name == kernel_info_dict["name"]
    assert kernel_info.last_activity == kernel_info_dict["last_activity"]
    assert kernel_info.execution_state == kernel_info_dict['execution_state']
    assert kernel_info.connections == kernel_info_dict["connections"]
    assert kernel_info == KernelInfo(**kernel_info_dict)


def test_notebook_session_class(notebook_session, session_dict):
    assert notebook_session.path == session_dict["path"]
    assert notebook_session.name == session_dict["name"]
    assert notebook_session.type == session_dict['type']
    assert notebook_session.kernel == KernelInfo(**session_dict["kernel"])
    assert notebook_session.notebook == session_dict["notebook"]
    assert notebook_session == NotebookSession(**session_dict)


@pytest.mark.parametrize(
    "connection_file, expected_kernel_id",
    [
        (
            "kernel-f92b7c8b-0858-4d10-903c-b0631540fb36.json",
            "f92b7c8b-0858-4d10-903c-b0631540fb36",
        ),
        (
            "kernel-ee2b7c8b-0858-4d10-903c-b0631540fb36.json",
            "ee2b7c8b-0858-4d10-903c-b0631540fb36",
        ),
    ],
)
def test_extract_kernel_id(connection_file, expected_kernel_id):
    assert extract_kernel_id(connection_file) == expected_kernel_id



import json

from unittest.mock import Mock

import pytest
import nbformat

from jinja2 import Environment
from notebook.services.contents.filemanager import FileContentsManager
from tornado.testing import AsyncTestCase, gen_test
from tornado.web import Application, HTTPError
from tornado.httpserver import HTTPRequest
from traitlets.config import Config


from bookstore.clone import (
    build_notebook_model,
    build_file_model,
    BookstoreCloneHandler,
    BookstoreCloneAPIHandler,
)


def test_build_notebook_model():
    content = nbformat.v4.new_notebook()
    expected = {
        "type": "notebook",
        "format": "json",
        "content": content,
        "name": "my_notebook_name.ipynb",
        "path": "test_directory/my_notebook_name.ipynb",
    }
    path = "./test_directory/my_notebook_name.ipynb"
    nb_content = nbformat.writes(content)
    assert build_notebook_model(nb_content, path) == expected


def test_build_file_model():
    content = "my fancy file"
    expected = {
        "type": "file",
        "format": "text",
        "content": content,
        "name": "file_name.txt",
        "path": "test_directory/file_name.txt",
    }
    path = "./test_directory/file_name.txt"
    assert build_file_model(content, path) == expected


class TestCloneHandler(AsyncTestCase):
    def setUp(self):
        super().setUp()
        self.mock_application = Mock(
            spec=Application, ui_methods={}, ui_modules={}, settings={'jinja2_env': Environment()}
        )

    def get_handler(self, uri, app=None):
        if app is None:
            app = self.mock_application
        connection = Mock(context=Mock(protocol="https"))
        payload_request = HTTPRequest(
            method='GET',
            uri=uri,
            headers={"Host": "localhost:8888"},
            body=None,
            connection=connection,
        )
        return BookstoreCloneHandler(app, payload_request)

    @gen_test
    async def test_get_no_param(self):
        empty_handler = self.get_handler('/bookstore/clone')
        with pytest.raises(HTTPError):
            await empty_handler.get()

    @gen_test
    async def test_get_no_bucket(self):
        no_bucket_handler = self.get_handler('/bookstore/clone?s3_bucket=&s3_key=hi')
        with pytest.raises(HTTPError):
            await no_bucket_handler.get()

    @gen_test
    async def test_get_no_object_key(self):
        no_object_key_handler = self.get_handler('/bookstore/clone?s3_bucket=hello&s3_key=')
        with pytest.raises(HTTPError):
            await no_object_key_handler.get()

    @gen_test
    async def test_get_success(self):
        success_handler = self.get_handler('/bookstore/clone?s3_bucket=hello&s3_key=my_key')
        await success_handler.get()

    def test_gen_template_params(self):
        success_handler = self.get_handler('/bookstore/clone?s3_bucket=hello&s3_key=my_key')
        expected = {
            's3_bucket': 'hello',
            's3_key': 'my_key',
            'clone_api_url': 'https://localhost:8888/api/bookstore/clone',
            'redirect_contents_url': 'https://localhost:8888',
        }
        success_handler = self.get_handler('/bookstore/clone?s3_bucket=hello&s3_key=my_key')
        output = success_handler.construct_template_params(
            s3_bucket="hello", s3_object_key="my_key"
        )
        assert expected == output

    def test_gen_template_params_base_url(self):
        base_url_list = ['/my_base_url', '/my_base_url/', 'my_base_url/', 'my_base_url']
        expected = {
            's3_bucket': 'hello',
            's3_key': 'my_key',
            'clone_api_url': 'https://localhost:8888/my_base_url/api/bookstore/clone',
            'redirect_contents_url': 'https://localhost:8888',
        }
        for base_url in base_url_list:
            mock_app = Mock(
                spec=Application,
                ui_methods={},
                ui_modules={},
                settings={'jinja2_env': Environment(), "base_url": base_url},
            )

            success_handler = self.get_handler(
                '/bookstore/clone?s3_bucket=hello&s3_key=my_key', app=mock_app
            )
            output = success_handler.construct_template_params(
                s3_bucket="hello", s3_object_key="my_key"
            )
            assert expected == output


class TestCloneAPIHandler(AsyncTestCase):
    def setUp(self):
        super().setUp()
        mock_settings = {
            "BookstoreSettings": {
                "s3_access_key_id": "mock_id",
                "s3_secret_access_key": "mock_access",
            }
        }
        config = Config(mock_settings)

        self.mock_application = Mock(
            spec=Application,
            ui_methods={},
            ui_modules={},
            settings={
                'jinja2_env': Environment(),
                "config": config,
                "contents_manager": FileContentsManager(),
            },
        )

    def post_handler(self, body_dict, app=None):
        if app is None:
            app = self.mock_application
        body = json.dumps(body_dict).encode('utf-8')
        payload_request = HTTPRequest(
            method='POST', uri="/api/bookstore/cloned", headers=None, body=body, connection=Mock()
        )
        return BookstoreCloneAPIHandler(app, payload_request)

    @gen_test
    async def test_post_no_body(self):
        post_body_dict = {}
        empty_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await empty_handler.post()

    @gen_test
    async def test_post_empty_bucket(self):
        post_body_dict = {"s3_key": "my_key", "s3_bucket": ""}
        empty_bucket_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await empty_bucket_handler.post()

    @gen_test
    async def test_post_empty_key(self):
        post_body_dict = {"s3_key": "", "s3_bucket": "my_bucket"}
        empty_key_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await empty_key_handler.post()

    @gen_test
    async def test_post_nonsense_params(self):
        post_body_dict = {"s3_key": "my_key", "s3_bucket": "my_bucket"}
        success_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await success_handler.post()

    @gen_test
    async def test_private_clone_nonsense_params(self):
        s3_bucket = "my_key"
        s3_object_key = "my_bucket"
        post_body_dict = {"s3_key": s3_object_key, "s3_bucket": s3_bucket}
        success_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await success_handler._clone(s3_bucket, s3_object_key)

    def test_build_post_response_model(self):
        content = "some arbitrary content"
        expected = {
            "type": "file",
            "format": "text",
            "content": content,
            "name": "file_name.txt",
            "path": "test_directory/file_name.txt",
            "s3_path": "s3://my_bucket/original_key/may_be_different_than_storage.txt",
            'versionID': "eeee222eee",
        }

        s3_bucket = "my_bucket"
        s3_object_key = "original_key/may_be_different_than_storage.txt"
        obj = {'VersionId': "eeee222eee"}
        model = {
            "type": "file",
            "format": "text",
            "content": content,
            "name": "file_name.txt",
            "path": "test_directory/file_name.txt",
        }
        handler = self.post_handler({})
        actual = handler.build_post_response_model(model, obj, s3_bucket, s3_object_key)
        assert actual == expected

    @gen_test
    async def test_build_text_content_model(self):
        content = "some content"
        expected = {
            "type": "file",
            "format": "text",
            "content": content,
            "name": "file_name.txt",
            "path": "test_directory/file_name.txt",
        }

        class MyFakeClass:
            def __init__(self):
                pass

            async def read(self):
                return content.encode('utf-8')

        obj = {'Body': MyFakeClass()}
        path = "test_directory/file_name.txt"
        success_handler = self.post_handler({})
        model = await success_handler.build_content_model(obj, path)
        assert model == expected

    @gen_test
    async def test_build_notebook_content_model(self):
        content = nbformat.v4.new_notebook()
        expected = {
            "type": "notebook",
            "format": "json",
            "content": content,
            "name": "file_name.ipynb",
            "path": "test_directory/file_name.ipynb",
        }

        class MyFakeClass:
            def __init__(self):
                pass

            async def read(self):
                return nbformat.writes(content).encode('utf-8')

        obj = {'Body': MyFakeClass()}
        path = "test_directory/file_name.ipynb"
        success_handler = self.post_handler({})
        model = await success_handler.build_content_model(obj, path)
        assert model == expected



import json

import aiobotocore

from botocore.exceptions import ClientError
from nbformat import ValidationError
from nbformat import validate as validate_nb
from notebook.base.handlers import APIHandler, path_regex
from notebook.services.contents.handlers import validate_model
from tornado import web

from .bookstore_config import BookstoreSettings
from .s3_paths import s3_path
from .s3_paths import s3_key
from .s3_paths import s3_display_path
from .utils import url_path_join


class BookstorePublishAPIHandler(APIHandler):
    """Publish a notebook to the publish path"""

    def initialize(self):
        """Initialize a helper to get bookstore settings and session information quickly"""
        self.bookstore_settings = BookstoreSettings(config=self.config)
        self.session = aiobotocore.get_session()

    @web.authenticated
    async def put(self, path):
        """Publish a notebook on a given path.

        PUT /api/bookstore/publish

        The payload directly matches the contents API for PUT.

        Parameters
        ----------
        path: str
            Path describing where contents should be published to, postfixed to the published_prefix .
        """
        if path == '' or path == '/':
            raise web.HTTPError(400, "Must provide a path for publishing")
        path = path.lstrip('/')

        s3_object_key = s3_key(self.bookstore_settings.published_prefix, path)

        model = self.get_json_body()
        self.validate_model(model)

        full_s3_path = s3_display_path(
            self.bookstore_settings.s3_bucket, self.bookstore_settings.published_prefix, path
        )
        self.log.info(f"Publishing to {full_s3_path}")

        obj = await self._publish(model['content'], s3_object_key)
        resp_content = self.prepare_response(obj, full_s3_path)

        self.set_status(obj['ResponseMetadata']['HTTPStatusCode'])
        self.finish(json.dumps(resp_content))

    def validate_model(self, model):
        """Checks that the model given to the API handler meets bookstore's expected structure for a notebook.

        Pattern for surfacing nbformat validation errors originally written in
        https://github.com/jupyter/notebook/blob/a44a367c219b60a19bee003877d32c3ff1ce2412/notebook/services/contents/manager.py#L353-L355

        Parameters
        ----------
        model: dict
            Request model for publishing describing the type and content of the object.

        Raises
        ------
        tornado.web.HTTPError
            Your model does not validate correctly
        """
        if not model:
            raise web.HTTPError(400, "Bookstore cannot publish an empty model")
        if model.get('type', "") != 'notebook':
            raise web.HTTPError(415, "Bookstore only publishes notebooks")

        content = model.get('content', {})
        if content == {}:
            raise web.HTTPError(422, "Bookstore cannot publish empty contents")
        try:
            validate_nb(content)
        except ValidationError as e:
            raise web.HTTPError(
                422,
                "Bookstore cannot publish invalid notebook. "
                "Validation errors are as follows: "
                f"{e.message} {json.dumps(e.instance, indent=1, default=lambda obj: '<UNKNOWN>')}",
            )

    async def _publish(self, content, s3_object_key):
        """Publish notebook model to the path
        
        Returns
        --------
        dict
            S3 PutObject response object
        """

        async with self.session.create_client(
            's3',
            aws_secret_access_key=self.bookstore_settings.s3_secret_access_key,
            aws_access_key_id=self.bookstore_settings.s3_access_key_id,
            endpoint_url=self.bookstore_settings.s3_endpoint_url,
            region_name=self.bookstore_settings.s3_region_name,
        ) as client:
            self.log.info(f"Processing published write to {s3_object_key}")
            try:
                obj = await client.put_object(
                    Bucket=self.bookstore_settings.s3_bucket,
                    Key=s3_object_key,
                    Body=json.dumps(content),
                )
            except ClientError as e:
                status_code = e.response['ResponseMetadata'].get('HTTPStatusCode')
                raise web.HTTPError(status_code, e.args[0])
            self.log.info(f"Done with published write to {s3_object_key}")

        return obj

    def prepare_response(self, obj, full_s3_path):
        """Prepares repsonse to publish PUT request.

        Parameters
        ----------
        obj: dict
            Validation dictionary for determining which endpoints to enable.
        path: 
            path to place after the published prefix in the designated bucket

        Returns
        --------
        dict
            Model for responding to put request. 
        """

        resp_content = {"s3_path": full_s3_path}

        if 'VersionId' in obj:
            resp_content["versionID"] = obj['VersionId']

        return resp_content



import pytest

from bookstore.utils import url_path_join


def test_url_join_pieces():
    bucket = 'mybucket'
    prefix = 'yo'
    path = 'pickles'
    assert url_path_join(bucket, prefix, path) == 'mybucket/yo/pickles'


def test_url_join_no_pieces():
    with pytest.raises(IndexError):
        url_path_join()


def test_url_join_one_piece():
    assert url_path_join('mypath') == 'mypath'


def test_url_path_join_strip_slash():
    assert url_path_join('/bucket/', 'yo', 'pickles/') == '/bucket/yo/pickles/'
    assert url_path_join('bucket/', '/yo/', '/pickles') == 'bucket/yo/pickles'
    assert url_path_join('/bucket/', '/yo/', '/pickles/') == '/bucket/yo/pickles/'
    assert url_path_join('/', '/') == '/'



"""Archival of notebooks"""

import json
from asyncio import Lock
from typing import Dict
from typing import NamedTuple

import aiobotocore
from notebook.services.contents.filemanager import FileContentsManager
from tornado import ioloop

from .bookstore_config import BookstoreSettings
from .s3_paths import s3_key, s3_display_path


class ArchiveRecord(NamedTuple):
    """Represents an archival record.

    An `ArchiveRecord` uses a Typed version of `collections.namedtuple()`. The
    record is immutable.

    Example
    -------

    An archive record (`filepath`, `content`, `queued_time`) contains:

    - a `filepath` to the record
    - the `content` for archival
    - the `queued time` length of time waiting in the queue for archiving
    """

    filepath: str
    content: str
    queued_time: float  # TODO: refactor to a datetime time


class BookstoreContentsArchiver(FileContentsManager):
    """Manages archival of notebooks to storage (S3) when notebook save occurs.

    This class is a custom Jupyter
    `FileContentsManager <https://jupyter-notebook.readthedocs.io/en/stable/extending/contents.html#contents-api>`_
    which holds information on storage location, path to it, and file to be
    written.

    Example
    -------

    - Bookstore settings combine with the parent Jupyter application settings.
    - A session is created for the current event loop.
    - To write to a particular path on S3, acquire a lock.
    - After acquiring the lock, `archive` method authenticates using the storage
      service's credentials.
    - If allowed, the notebook is queued to be written to storage (i.e. S3).

    Attributes
    ----------

    path_locks : dict
        Dictionary of paths to storage and the lock associated with a path.
    path_lock_ready: asyncio mutex lock
        A mutex lock associated with a path.
    """

    def __init__(self, *args, **kwargs):
        super(FileContentsManager, self).__init__(*args, **kwargs)

        # opt ourselves into being part of the Jupyter App that should have Bookstore Settings applied
        self.settings = BookstoreSettings(parent=self)

        self.log.info(
            "Archiving notebooks to {}".format(
                s3_display_path(self.settings.s3_bucket, self.settings.workspace_prefix)
            )
        )

        try:
            # create a session object from the current event loop
            self.session = aiobotocore.get_session()
        except Exception:
            self.log.warn("Unable to create a session")
            raise

        # a collection of locks per path to suppress writing while the path may be in use
        self.path_locks: Dict[str, Lock] = {}
        self.path_lock_ready = Lock()

    async def archive(self, record: ArchiveRecord):
        """Process a record to write to storage.

        Acquire a path lock before archive. Writing to storage will only be
        allowed to a path if a valid `path_lock` is held and the path is not
        locked by another process.

        Parameters
        ----------

        record : ArchiveRecord
            A notebook and where it should be written to storage
        """
        async with self.path_lock_ready:
            lock = self.path_locks.get(record.filepath)

            if lock is None:
                lock = Lock()
                self.path_locks[record.filepath] = lock

        # Skip writes when a given path is already locked
        if lock.locked():
            self.log.info("Skipping archive of %s", record.filepath)
            return

        async with lock:
            try:
                async with self.session.create_client(
                    's3',
                    aws_secret_access_key=self.settings.s3_secret_access_key,
                    aws_access_key_id=self.settings.s3_access_key_id,
                    endpoint_url=self.settings.s3_endpoint_url,
                    region_name=self.settings.s3_region_name,
                ) as client:
                    self.log.info("Processing storage write of %s", record.filepath)
                    file_key = s3_key(self.settings.workspace_prefix, record.filepath)
                    await client.put_object(
                        Bucket=self.settings.s3_bucket, Key=file_key, Body=record.content
                    )
                    self.log.info("Done with storage write of %s", record.filepath)
            except Exception as e:
                self.log.error(
                    'Error while archiving file: %s %s', record.filepath, e, exc_info=True
                )

    def run_pre_save_hook(self, model, path, **kwargs):
        """Send request to store notebook to S3.

        This hook offloads the storage request to the event loop.
        When the event loop is available for execution of the request, the
        storage of the notebook will be done and the write to storage occurs.

        Parameters
        ----------

        model : dict
            The type of file and its contents
        path : str
            The storage location
        """
        if model["type"] != "notebook":
            self.log.debug(
                "Bookstore only archives notebooks, "
                f"request does not state that {path} is a notebook."
            )
            return

        content = json.dumps(model["content"])

        loop = ioloop.IOLoop.current()

        # Offload archival and schedule write to storage with the current event loop
        loop.spawn_callback(
            self.archive,
            ArchiveRecord(
                content=content, filepath=path, queued_time=ioloop.IOLoop.current().time()
            ),
        )



"""Handler to clone notebook from storage."""
import json
import os

from copy import deepcopy

import aiobotocore

from botocore.exceptions import ClientError
from jinja2 import FileSystemLoader
from notebook.base.handlers import IPythonHandler, APIHandler
from tornado import web

from . import PACKAGE_DIR
from .bookstore_config import BookstoreSettings
from .s3_paths import s3_path, s3_display_path
from .utils import url_path_join

BOOKSTORE_FILE_LOADER = FileSystemLoader(PACKAGE_DIR)


def build_notebook_model(content, path):
    """Helper that builds a Contents API compatible model for notebooks.

    Parameters
    ----------
    content : str
        The content of the model.
    path : str
        The path to be targeted.

    Returns
    --------
    dict
        Jupyter Contents API compatible model for notebooks
    """
    model = {
        "type": "notebook",
        "format": "json",
        "content": json.loads(content),
        "name": os.path.basename(os.path.relpath(path)),
        "path": os.path.relpath(path),
    }
    return model


def build_file_model(content, path):
    """Helper that builds a Contents API compatible model for files.

    Parameters
    ----------
    content: str
        The content of the model
    path : str
        The path to be targeted.

    Returns
    --------
    dict
        Jupyter Contents API compatible model for files
    """
    model = {
        "type": "file",
        "format": "text",
        "content": content,
        "name": os.path.basename(os.path.relpath(path)),
        "path": os.path.relpath(path),
    }
    return model


class BookstoreCloneHandler(IPythonHandler):
    """Prepares and provides clone options page, populating UI with clone option parameters.

    Provides handling for ``GET`` requests when cloning a notebook
    from storage (S3). Launches a user interface with cloning options.

    Methods
    -------
    initialize(self)
        Helper to access bookstore settings.
    get(self)
        Checks for valid storage settings and render a UI for clone options.
    construct_template_params(self, s3_bucket, s3_object_key)
        Helper to populate Jinja template for cloning option page.
    get_template(self, name)
        Loads a Jinja template and its related settings.

    See also
    --------
    `Jupyter Notebook reference on Custom Handlers <https://jupyter-notebook.readthedocs.io/en/stable/extending/handlers.html#registering-custom-handlers>`_
    """

    def initialize(self):
        """Helper to retrieve bookstore setting for the session."""
        self.bookstore_settings = BookstoreSettings(config=self.config)

    @web.authenticated
    async def get(self):
        """GET /bookstore/clone?s3_bucket=<your_s3_bucket>&s3_key=<your_s3_key>

        Renders an options page that will allow you to clone a notebook
        from a specific bucket via the Bookstore cloning API.

        s3_bucket is the bucket you wish to clone from.
        s3_key is the object key that you wish to clone.
        """
        s3_bucket = self.get_argument("s3_bucket")
        if s3_bucket == '' or s3_bucket == "/":
            raise web.HTTPError(400, "Requires an S3 bucket in order to clone")

        # s3_paths module has an s3_key function; s3_object_key avoids confusion
        s3_object_key = self.get_argument("s3_key")
        if s3_object_key == '' or s3_object_key == '/':
            raise web.HTTPError(400, "Requires an S3 object key in order to clone")

        self.log.info(f"Setting up cloning landing page for {s3_object_key}")

        template_params = self.construct_template_params(s3_bucket, s3_object_key)
        self.set_header('Content-Type', 'text/html')
        self.write(self.render_template('clone.html', **template_params))

    def construct_template_params(self, s3_bucket, s3_object_key):
        """Helper that takes valid S3 parameters and populates UI template
        
        Returns
        --------
        
        dict
            Template parameters in a dictionary
        """
        base_uri = f"{self.request.protocol}://{self.request.host}"
        clone_api_url = url_path_join(base_uri, self.base_url, "/api/bookstore/clone")
        redirect_contents_url = url_path_join(base_uri, self.default_url)
        template_params = {
            "s3_bucket": s3_bucket,
            "s3_key": s3_object_key,
            "clone_api_url": clone_api_url,
            "redirect_contents_url": redirect_contents_url,
        }
        return template_params

    def get_template(self, name):
        """Loads a Jinja template by name."""
        return BOOKSTORE_FILE_LOADER.load(self.settings['jinja2_env'], name)


class BookstoreCloneAPIHandler(APIHandler):
    """Handle notebook clone from storage.

    Provides API handling for ``POST`` and clones a notebook
    from storage (S3).

    Methods
    -------
    initialize(self)
        Helper to access bookstore settings.
    post(self)
        Clone a notebook from the location specified by the payload.
    build_content_model(self, obj, path)
        Helper that takes a response from S3 and creates a ContentsAPI compatible model.
    build_post_response_model(self, model, obj, s3_bucket, s3_object_key)
        Helper that takes a Jupyter Contents API compliant model and adds cloning specific information.

    See also
    --------
    `Jupyter Notebook reference on Custom Handlers <https://jupyter-notebook.readthedocs.io/en/stable/extending/handlers.html#registering-custom-handlers>`_
    """

    def initialize(self):
        """Helper to retrieve bookstore setting for the session."""
        self.bookstore_settings = BookstoreSettings(config=self.config)

        self.session = aiobotocore.get_session()

    async def _clone(self, s3_bucket, s3_object_key):
        """Main function that handles communicating with S3 to initiate the clone.

        Parameters
        ----------
        s3_bucket: str
            Log (usually from the NotebookApp) for logging endpoint changes.
        s3_object_key: str
            The the path we wish to clone to.
        """

        self.log.info(f"bucket: {s3_bucket}")
        self.log.info(f"key: {s3_object_key}")

        async with self.session.create_client(
            's3',
            aws_secret_access_key=self.bookstore_settings.s3_secret_access_key,
            aws_access_key_id=self.bookstore_settings.s3_access_key_id,
            endpoint_url=self.bookstore_settings.s3_endpoint_url,
            region_name=self.bookstore_settings.s3_region_name,
        ) as client:
            self.log.info(f"Processing clone of {s3_object_key}")
            try:
                obj = await client.get_object(Bucket=s3_bucket, Key=s3_object_key)
            except ClientError as e:
                status_code = e.response['ResponseMetadata'].get('HTTPStatusCode')
                raise web.HTTPError(status_code, e.args[0])

            self.log.info(f"Obtained contents for {s3_object_key}")

        return obj

    @web.authenticated
    async def post(self):
        """POST /api/bookstore/clone

        Clone a notebook to the path specified in the payload.

        The payload type for the request should be::

            {
            "s3_bucket": string,
            "s3_key": string,
            "target_path"?: string
            }

        The response payload should match the standard Jupyter contents
        API POST response.
        """
        model = self.get_json_body()
        s3_bucket = model.get("s3_bucket", "")
        if s3_bucket == '' or s3_bucket == "/":
            raise web.HTTPError(400, "Must have a bucket to clone from")

        # s3_paths module has an s3_key function; s3_object_key avoids confusion
        s3_object_key = model.get("s3_key", "")
        if s3_object_key == '' or s3_object_key == '/':
            raise web.HTTPError(400, "Must have a key to clone from")

        target_path = model.get("target_path", "") or os.path.basename(
            os.path.relpath(s3_object_key)
        )

        self.log.info(f"About to clone from {s3_object_key}")
        obj = await self._clone(s3_bucket, s3_object_key)

        content_model = await self.build_content_model(obj, target_path)

        self.log.info(f"Completing clone for {s3_object_key}")
        self.contents_manager.save(content_model, content_model['path'])

        resp_model = self.build_post_response_model(content_model, obj, s3_bucket, s3_object_key)

        self.set_status(obj['ResponseMetadata']['HTTPStatusCode'])
        self.set_header('Content-Type', 'application/json')
        self.finish(resp_model)

    async def build_content_model(self, obj, target_path):
        """Helper that takes a response from S3 and creates a ContentsAPI compatible model.
        
        If the file at target_path already exists, this increments the file name.
        
        Parameters
        ----------
        obj : dict
            Response object from S3
        target_path : str
            The the path we wish to clone to, may be incremented if already present.

        Returns
        --------
        dict
            Jupyter Contents API compatible model
        """
        path = self.contents_manager.increment_filename(target_path)
        content = await obj['Body'].read()
        content = content.decode('utf-8')
        if os.path.splitext(path)[1] in [".ipynb", ".jpynb"]:
            model = build_notebook_model(content, path)
        else:
            model = build_file_model(content, path)
        return model

    def build_post_response_model(self, model, obj, s3_bucket, s3_object_key):
        """Helper that takes a Jupyter Contents API compliant model and adds cloning specific information.

        Parameters
        ----------
        model : dict
            Jupyter Contents API model
        obj : dict
            Log (usually from the NotebookApp) for logging endpoint changes.
        s3_bucket : str
            The S3 bucket we are cloning from
        s3_object_key: str
            The S3 key we are cloning

        Returns
        --------
        dict
            Model with additional info about the S3 cloning
        """
        model = deepcopy(model)
        model["s3_path"] = s3_display_path(s3_bucket, s3_object_key)
        if 'VersionId' in obj:
            model["versionID"] = obj['VersionId']
        return model



"""Client to interact with a notebook's bookstore functionality."""
import requests

from .nb_client import CurrentNotebookClient


class BookstoreClient(CurrentNotebookClient):
    """EXPERIMENTAL SUPPORT: A client that allows access to a Bookstore from within a notebook.
    
    Parameters
    ----------
    s3_bucket: str
        (optional) Provide a default bucket for this bookstore client to clone from.
    

    Attributes
    ----------
    default_bucket : str
        The default bucket to be used for cloning.
    """

    def __init__(self, s3_bucket=None):
        if s3_bucket:
            self.default_bucket = s3_bucket
        super().__init__()

    @property
    def publish_endpoint(self):
        """Helper to refer to construct the publish endpoint for this notebook server."""
        api_endpoint = "/api/bookstore/publish/"
        return f"{self.url}{api_endpoint}"

    def publish(self, path=None):
        """Publish notebook to bookstore
        
        Parameters
        ----------
        path : str
            (optional) Path that you wish to publish; defaults to current notebook.
        s3_object_key: str
            The the path we wish to clone to.
        """
        if path is None:
            path = self.session.path
        nb_json = self.get_contents(path)['content']
        json_body = {"type": "notebook", "content": nb_json}

        target_url = f"{self.publish_endpoint}{path}"

        response = self.req_session.put(target_url, json=json_body)
        return response

    @property
    def clone_endpoint(self):
        """Helper to refer to construct the clone endpoint for this notebook server."""
        api_endpoint = "/api/bookstore/clone/"
        return f"{self.url}{api_endpoint}"

    def clone(self, s3_bucket="", s3_key="", target_path=""):
        """Clone files via bookstore.
        
        Parameters
        ----------
        s3_bucket : str
            (optional) S3 bucket you wish to clone from; defaults to client's bucket.
        s3_object_key: str
            The object key describing the object you wish to clone from S3.
        target_path: str
            (optional) The location you wish to clone the object to; defaults to s3_object_key.
        """
        s3_bucket = s3_bucket or self.default_bucket
        json_body = {"s3_bucket": s3_bucket, "s3_key": s3_key, "target_path": target_path}
        # TODO: Add a check for success
        response = self.req_session.post(f"{self.clone_endpoint}", json=json_body)
        return response



""" Helper for interactive debug of cloning endpoint.

This module is intended to be used to interactively when debugging the
bookstore cloning endpoint.

To use and test this, you'll need to separately run a Jupyter server.

Example Usage
-------------

To test the 'nteract-notebooks' bucket on S3 you'll need to run a command like
the following::

jupyter notebook --NotebookApp.allow_origin="*" --NotebookApp.disable_check_xsrf=True \
    --no-browser --NotebookApp.token="" \
    --BookstoreSettings.s3_bucket=nteract-notebooks

To test the 'bookstore' bucket created with minio you'll need to run a command
like the following::

jupyter notebook --NotebookApp.allow_origin="*" --NotebookApp.disable_check_xsrf=True \
    --no-browser --NotebookApp.token="" \
    --BookstoreSettings.s3_endpoint_url="http://localhost:9000" \
    --BookstoreSettings.s3_access_key_id="ONLY_ON_CIRCLE" \
    --BookstoreSettings.s3_secret_access_key="CAN_WE_DO_THIS"

Additional examples
-------------------

See the docstring under the ``if __name__`` statement below.
"""
import nbformat
import pprint
import requests


def get(queries):
    return requests.get(f"http://localhost:8888/api/bookstore/cloned{queries}")


def post(**kwargs):
    return requests.post("http://localhost:8888/api/bookstore/cloned", json={**kwargs})


if __name__ == "__main__":
    """
    Examples
    --------

    # tests s3 hosted 'nteract-notebooks' bucket
    response = get("?s3_bucket=nteract-notebooks&s3_key=published/whateverwewant.json")
    print(response.content)
    response = get("/?s3_bucket=nteract-notebooks&s3_key=published/whateverwewant.json")
    print(response.content)
    response = get("?s3_bucket=nteract-notebooks&s3_key=Introduction_to_Chainer.ipynb")
    print(response.content)


    # tests minio created 'bookstore' bucket
    response = get("?s3_bucket=bookstore&s3_key=ci-published/ci-published.ipynb")
    print(response.content)
    response = post(s3_bucket="nteract-notebooks", s3_key="Introduction_to_Chainer.ipynb")
    print(response.content)

    Changing the requests will enable you to test different situations and
    analyze the responses.
    """
    response = post(s3_bucket="bookstore", s3_key="ci-published/ci-published.ipynb")
    pprint.pprint(response.json())
    response = post(
        s3_bucket="bookstore",
        s3_key="ci-published/ci-published.ipynb",
        target_path="published/this_is_my_target_path.ipynb",
    )
    pprint.pprint(response.json())



"""Utility and helper functions."""


def url_path_join(*pieces):
    """Join components into a relative url.
    
    Use to prevent double slash when joining subpath. This will leave the
    initial and final / in place.
    
    Code based on Jupyter notebook `url_path_join`.
    """
    initial = pieces[0].startswith('/')
    final = pieces[-1].endswith('/')
    stripped = [s.strip('/') for s in pieces]
    result = '/'.join(s for s in stripped if s)
    if initial:
        result = '/' + result
    if final:
        result = result + '/'
    if result == '//':
        result = '/'
    return result



import asyncio
import json

from unittest.mock import Mock

import pytest

from bookstore.publish import BookstorePublishAPIHandler
from nbformat.v4 import new_notebook
from tornado.testing import AsyncTestCase, gen_test
from tornado.web import Application, HTTPError
from tornado.httpserver import HTTPRequest
from traitlets.config import Config


def test_create_publish_handler_no_params():
    with pytest.raises(TypeError):
        assert BookstorePublishAPIHandler()


class TestPublishAPIHandler(AsyncTestCase):
    def setUp(self):
        super().setUp()
        mock_settings = {
            "BookstoreSettings": {
                "s3_access_key_id": "mock_id",
                "s3_secret_access_key": "mock_access",
                "s3_bucket": "my_bucket",
                "published_prefix": "custom_prefix",
            }
        }
        config = Config(mock_settings)

        self.mock_application = Mock(
            spec=Application, ui_methods={}, ui_modules={}, settings={"config": config}
        )

    def put_handler(self, uri, body_dict=None, app=None):
        if body_dict is None:
            body_dict = {}
        if app is None:
            app = self.mock_application
        connection = Mock(context=Mock(protocol="https"))
        body = json.dumps(body_dict).encode('utf-8')
        payload_request = HTTPRequest(
            method='PUT',
            uri=uri,
            headers={"Host": "localhost:8888"},
            body=body,
            connection=connection,
        )
        return BookstorePublishAPIHandler(app, payload_request)

    @gen_test
    async def test_put_no_path(self):
        no_path_handler = self.put_handler('/bookstore/publish/')
        with pytest.raises(HTTPError):
            await no_path_handler.put('')

    @gen_test
    async def test_put_no_body(self):
        no_body_handler = self.put_handler('/bookstore/publish/hi')
        with pytest.raises(HTTPError):
            await no_body_handler.put('hi')

    @gen_test
    async def test_put_s3_error(self):
        """this test includes a valid body so that we get to the s3 part of our system"""
        body_dict = {'content': new_notebook(), 'type': "notebook"}
        ok_body_handler = self.put_handler('/bookstore/publish/hi', body_dict=body_dict)
        with pytest.raises(HTTPError):
            await ok_body_handler.put('hi')

    def test_prepare_response(self):
        expected = {"s3_path": "s3://my_bucket/custom_prefix/mylocal/path", "versionID": "eeeeAB"}
        empty_handler = self.put_handler('/bookstore/publish/hi')
        actual = empty_handler.prepare_response(
            {"VersionId": "eeeeAB"}, "s3://my_bucket/custom_prefix/mylocal/path"
        )
        assert actual == expected

    def test_validate_model_no_type(self):
        body_dict = {'content': {}}
        empty_handler = self.put_handler('/bookstore/publish/hi')
        with pytest.raises(HTTPError):
            empty_handler.validate_model(body_dict)

    def test_validate_model_wrong_type(self):
        body_dict = {'content': {}, 'type': "file"}
        empty_handler = self.put_handler('/bookstore/publish/hi')
        with pytest.raises(HTTPError):
            empty_handler.validate_model(body_dict)

    def test_validate_model_empty_content(self):
        body_dict = {'content': {}, 'type': "notebook"}
        empty_handler = self.put_handler('/bookstore/publish/hi')
        with pytest.raises(HTTPError):
            empty_handler.validate_model(body_dict)

    def test_validate_model_bad_notebook(self):
        bad_notebook = new_notebook()
        bad_notebook['other_field'] = "hello"
        body_dict = {'content': bad_notebook, 'type': "notebook"}
        empty_handler = self.put_handler('/bookstore/publish/hi')
        with pytest.raises(HTTPError):
            empty_handler.validate_model(body_dict)

    def test_validate_model_good_notebook(self):
        body_dict = {'content': new_notebook(), 'type': "notebook"}
        empty_handler = self.put_handler('/bookstore/publish/hi')
        empty_handler.validate_model(body_dict)



"""Client for accessing notebook server endpoints from within a notebook.
"""
import json
import os
import re
from copy import deepcopy
from typing import NamedTuple

import requests
from IPython import get_ipython
from notebook.notebookapp import list_running_servers


def extract_kernel_id(connection_file):
    """Get the kernel id string from a file"""
    # regex is used as a more robust approach than lstrip
    connection_filename = os.path.basename(connection_file)
    kernel_id = re.sub(r"kernel-(.*)\.json", r"\1", connection_filename)
    return kernel_id


class LiveNotebookRecord(NamedTuple):
    """Representation of live notebook server.

    This is a record of an object returned by
    `notebook.notebookapp.list_running_servers()`.

    Example
    -------
    :: 

        [{'base_url': '/',
        'hostname': 'localhost',
        'notebook_dir': '/Users/mpacer/jupyter/eg_notebooks',
        'password': False,
        'pid': 96033,
        'port': 8888,
        'secure': False,
        'token': '',
        'url': 'http://localhost:8888/'}]

    """

    base_url: str
    hostname: str
    notebook_dir: str
    password: bool
    pid: int
    port: int
    secure: bool
    token: str
    url: str


class KernelInfo:
    """Representation of kernel info returned by the notebook's /api/kernel endpoint.

    Attributes
    ----------
    id: str
    name: str
    last_activity: str
    execution_state: str 
    connections: int

    Example
    -------
    ::

        {id: 'f92b7c8b-0858-4d10-903c-b0631540fb36',
        name: 'dev',
        last_activity: '2019-03-14T23:38:08.137987Z',
        execution_state: 'idle',
        connections: 0}
    """

    def __init__(self, *args, id, name, last_activity, execution_state, connections):
        self.model = {
            "id": id,
            "name": name,
            "last_activity": last_activity,
            "execution_state": execution_state,
            "connections": connections,
        }
        self.id = id
        self.name = name
        self.last_activity = last_activity
        self.execution_state = execution_state
        self.connections = connections

    def __repr__(self):
        return json.dumps(self.model, indent=2)

    def __eq__(self, other):
        if isinstance(other, KernelInfo):
            cmp_attrs = [
                self.id == other.id,
                self.name == other.name,
                self.last_activity == other.last_activity,
                self.execution_state == other.execution_state,
                self.connections == other.connections,
            ]
            return all(cmp_attrs)
        else:
            return False


class NotebookSession:
    """Representation of session info returned by the notebook's /api/sessions/ endpoint.
    
    Attributes
    ----------
    id: str
    path: str 
    name: str 
    type: str 
    kernel: KernelInfo
    notebook: dict 
    model: dict
        Record of the raw response (without converting the KernelInfo).

    Example
    -------
    ::
    
        {id: '68d9c58f-c57d-4133-8b41-5ec2731b268d',
         path: 'Untitled38.ipynb',
         name: '',
         type: 'notebook',
         kernel: KernelInfo(id='f92b7c8b-0858-4d10-903c-b0631540fb36', 
                            name='dev', 
                            last_activity='2019-03-14T23:38:08.137987Z', 
                            execution_state='idle', 
                            connections=0),
        notebook: {'path': 'Untitled38.ipynb', 'name': ''}}
    """

    def __init__(self, *args, path, name, type, kernel, notebook={}, **kwargs):
        self.model = {
            "path": path,
            "name": name,
            "type": type,
            "kernel": kernel,
            "notebook": notebook,
        }
        self.path = path
        self.name = name
        self.type = type
        self.kernel = KernelInfo(**kernel)
        self.notebook = notebook

    def __repr__(self):
        return json.dumps(self.model, indent=2)

    def __eq__(self, other):
        if isinstance(other, NotebookSession):
            cmp_attrs = [
                # self.id == other.id,
                self.path == other.path,
                self.name == other.name,
                self.type == other.type,
                self.kernel == other.kernel,
                self.notebook == other.notebook,
            ]
            return all(cmp_attrs)
        else:
            return False


class NotebookClient:
    """EXPERIMENTAL SUPPORT: Client used to interact with a notebook server from within a notebook.

    Parameters
    ----------
    nb_config: dict
        Dictionary of info compatible with creating a LiveNotebookRecord.

    Attributes
    ----------
    nb_config: dict
       Dictionary of info compatible with creating a LiveNotebookRecord. 
    nb_record: LiveNotebookRecord
        LiveNotebookRecord of info for this notebook
    url: str
        url from nb_record minus final /
    token: str
        token used for authenticating requests serverside
    xsrf_token: str
        xsrf_token used in cookie for authenticating requests
    req_session: requests.Session
        Session to be reused across methods
    """

    def __init__(self, nb_config):
        self.nb_config = nb_config
        self.nb_record = LiveNotebookRecord(**self.nb_config)
        self.url = self.nb_record.url.rstrip(
            "/"
        )  # So that we can have full API endpoints without double //
        self.setup_auth()
        self.setup_request_sessions()

    def setup_auth(self):
        """ Sets up token access for authorizing requests to notebook server.

        This sets the notebook token as self.token and the xsrf_token as self.xsrf_token.
        """
        self.token = self.nb_record.token
        first = requests.get(f"{self.url}/login")
        self.xsrf_token = first.cookies.get("_xsrf", "")

    def setup_request_sessions(self):
        """ Sets up a requests.Session object for sharing headers across API requests. """
        self.req_session = requests.Session()
        self.req_session.headers.update(self.headers)

    @property
    def sessions(self):
        """Current notebook sessions. Reissues request on each call. """
        return {
            session['kernel']['id']: NotebookSession(**session) for session in self.get_sessions()
        }

    @property
    def headers(self):
        """Default headers to be shared across requests. """
        headers = {
            'Authorization': f'token {self.token}',
            'X-XSRFToken': self.xsrf_token,
            "Content-Type": "application/json",
        }
        return headers

    @property
    def sessions_endpoint(self):
        """Current server's kernels API endpoint."""
        api_endpoint = "/api/sessions/"
        return f"{self.url}{api_endpoint}"

    def get_sessions(self):
        """Requests info about current sessions from notebook server."""
        target_url = f"{self.sessions_endpoint}"
        resp = self.req_session.get(target_url)
        return resp.json()

    @property
    def kernels_endpoint(self):
        """Current server's kernels API endpoint."""
        api_endpoint = "/api/kernels/"
        return f"{self.url}{api_endpoint}"

    def get_kernels(self):
        """Requests info about current kernels from notebook server."""
        target_url = f"{self.kernels_endpoint}"
        resp = self.req_session.get(target_url)
        return resp.json()

    @property
    def kernels(self):
        """Current notebook kernels. Reissues request on each call."""
        return self.get_kernels()

    @property
    def contents_endpoint(self):
        """Current server's contents API endpoint."""
        api_endpoint = "/api/contents/"
        return f"{self.url}{api_endpoint}"

    def get_contents(self, path):
        """Requests info about current contents from notebook server."""
        target_url = f"{self.contents_endpoint}{path}"
        resp = self.req_session.get(target_url)
        return resp.json()


class NotebookClientCollection:
    """EXPERIMENTAL SUPPORT: Representation of a collection of notebook clients"""

    # TODO: refactor from lambda to a def
    nb_client_gen = lambda: (NotebookClient(x) for x in list_running_servers())
    sessions = {x.url: x.sessions for x in nb_client_gen()}

    @classmethod
    def current_server(cls):
        """class method for current notebook server"""

        current_kernel_id = extract_kernel_id(get_ipython().parent.parent.connection_file)
        for server_url, session_dict in cls.sessions.items():
            for session_id, session in session_dict.items():
                if session.kernel.id == current_kernel_id:
                    return next(
                        client for client in cls.nb_client_gen() if client.url == server_url
                    )


class CurrentNotebookClient(NotebookClient):
    """EXPERIMENTAL SUPPORT: Represents the currently active notebook client."""

    def __init__(self):
        self.nb_client = NotebookClientCollection.current_server()
        super().__init__(self.nb_client.nb_config)
        self.session = self.sessions[self.kernel_id]

    @property
    def connection_file(self):
        """Connection file for connecting to current notebook's kernel."""
        return get_ipython().parent.parent.connection_file

    @property
    def kernel_id(self):
        """Kernel id for identifying which notebook is currently being used by this session."""
        return extract_kernel_id(self.connection_file)



"""Configuration settings for bookstore."""
from traitlets import Integer, Unicode, Bool
from traitlets.config import LoggingConfigurable


class BookstoreSettings(LoggingConfigurable):
    """Configuration for archival and publishing.

    Settings include storage directory locations, S3 authentication,
    additional S3 settings, and Bookstore resources.

    S3 authentication settings can be set, or they can be left unset when
    IAM is used.

    Like the Jupyter notebook, bookstore uses traitlets to handle
    configuration, loading from files or CLI.

    Attributes
    ----------
    workspace_prefix : str(``workspace``)
                       Directory to use for user workspace storage
    published_prefix : str(``published``)
                       Directory to use for published notebook storage
    s3_access_key_id : str, optional
                       Environment variable ``JPYNB_S3_ACCESS_KEY_ID``
    s3_secret_access_key : str, optional
                       Environment variable ``JPYNB_S3_SECRET_ACCESS_KEY``
    s3_endpoint_url : str(``"https://s3.amazonaws.com"``)
                       Environment variable ``JPYNB_S3_ENDPOINT_URL``
    s3_region_name : str(``"us-east-1"``)
                       Environment variable ``JPYNB_S3_REGION_NAME``
    s3_bucket : str(``""``)
                Bucket name, environment variable ``JPYNB_S3_BUCKET``
    max_threads : int(``16``)
                  Maximum threads from the threadpool available for S3 read/writes
    """

    workspace_prefix = Unicode("workspace", help="Prefix for the live workspace notebooks").tag(
        config=True
    )
    published_prefix = Unicode("published", help="Prefix for published notebooks").tag(config=True)
    enable_cloning = Bool(True, help="Enable cloning.").tag(config=True)

    s3_access_key_id = Unicode(
        help="S3/AWS access key ID", allow_none=True, default_value=None
    ).tag(config=True, env="JPYNB_S3_ACCESS_KEY_ID")
    s3_secret_access_key = Unicode(
        help="S3/AWS secret access key", allow_none=True, default_value=None
    ).tag(config=True, env="JPYNB_S3_SECRET_ACCESS_KEY")

    s3_endpoint_url = Unicode("https://s3.amazonaws.com", help="S3 endpoint URL").tag(
        config=True, env="JPYNB_S3_ENDPOINT_URL"
    )
    s3_region_name = Unicode("us-east-1", help="Region name").tag(
        config=True, env="JPYNB_S3_REGION_NAME"
    )
    s3_bucket = Unicode("", help="Bucket name to store notebooks").tag(
        config=True, env="JPYNB_S3_BUCKET"
    )

    max_threads = Integer(
        16, help="Maximum number of threads for the threadpool allocated for S3 read/writes"
    ).tag(config=True)


def validate_bookstore(settings: BookstoreSettings):
    """Check that settings exist.

    Parameters
    ----------
    settings : bookstore.bookstore_config.BookstoreSettings
               Instantiated settings object to be validated.

    Returns
    -------
    validation_checks : dict
        Statements about whether features are validly configured and available
    """
    general_settings = [settings.s3_bucket != "", settings.s3_endpoint_url != ""]
    archive_settings = [*general_settings, settings.workspace_prefix != ""]
    published_settings = [*general_settings, settings.published_prefix != ""]
    cloning_settings = [settings.enable_cloning]

    validation_checks = {
        "bookstore_valid": all(general_settings),
        "archive_valid": all(archive_settings),
        "publish_valid": all(published_settings),
        "clone_valid": all(cloning_settings),
    }
    return validation_checks



#!/usr/bin/env python
# -*- coding: utf-8 -*-

print("using CI config")

from bookstore import BookstoreContentsArchiver, BookstoreSettings

# jupyter config
# At ~/.jupyter/jupyter_notebook_config.py for user installs
# At __ for system installs
c = get_config()

c.NotebookApp.contents_manager_class = BookstoreContentsArchiver

c.BookstoreSettings.workspace_prefix = "ci-workspace"
c.BookstoreSettings.published_prefix = "ci-published"

# If using minio for development
c.BookstoreSettings.s3_endpoint_url = "http://localhost:9000"
c.BookstoreSettings.s3_bucket = "bookstore"

# Straight out of `circleci/config.yml`
c.BookstoreSettings.s3_access_key_id = "ONLY_ON_CIRCLE"
c.BookstoreSettings.s3_secret_access_key = "CAN_WE_DO_THIS"



"""Tests for archive"""
import asyncio
import pytest
import json
import logging

from bookstore.archive import ArchiveRecord, BookstoreContentsArchiver
from nbformat.v4 import new_notebook


def test_create_contentsarchiver():
    assert BookstoreContentsArchiver()


def test_create_contentsarchiver_invalid_args_count():
    with pytest.raises(TypeError):
        BookstoreContentsArchiver(42, True, 'hello')


@pytest.mark.asyncio
async def test_archive_failure_on_no_lock():
    archiver = BookstoreContentsArchiver()
    assert archiver

    record = ArchiveRecord('my_notebook_path.ipynb', json.dumps(new_notebook()), 100.2)
    assert record

    await archiver.archive(record)


@pytest.mark.asyncio
async def test_archive_abort_with_lock(caplog):
    """Acquire a lock in advance so that when the archiver attempts to archive, it will abort."""

    archiver = BookstoreContentsArchiver()
    record = ArchiveRecord('my_notebook_path.ipynb', json.dumps(new_notebook()), 100.2)

    lock = asyncio.Lock()
    archiver.path_locks['my_notebook_path.ipynb'] = lock
    async with lock:
        with caplog.at_level(logging.INFO):
            await archiver.archive(record)
    assert 'Skipping archive of my_notebook_path.ipynb' in caplog.text


def test_pre_save_hook():
    archiver = BookstoreContentsArchiver()
    model = {"type": "notebook", "content": new_notebook()}
    target_path = "my_notebook_path.ipynb"

    archiver.run_pre_save_hook(model, target_path)


def test_pre_save_hook_bad_model():
    archiver = BookstoreContentsArchiver()
    model = {"type": "file", "content": new_notebook()}
    target_path = "my_notebook_path.ipynb"

    archiver.run_pre_save_hook(model, target_path)



"""Tests for bookstore config"""
import pytest

from bookstore.bookstore_config import BookstoreSettings, validate_bookstore


def test_validate_bookstore_defaults():
    """Tests that all bookstore validates with default values."""
    expected = {
        "bookstore_valid": False,
        "publish_valid": False,
        "archive_valid": False,
        "clone_valid": True,
    }
    settings = BookstoreSettings()
    assert validate_bookstore(settings) == expected


def test_validate_bookstore_published():
    """Tests that bookstore does not validate with an empty published_prefix."""
    expected = {
        "bookstore_valid": True,
        "publish_valid": False,
        "archive_valid": True,
        "clone_valid": True,
    }
    settings = BookstoreSettings(s3_bucket="A_bucket", published_prefix="")
    assert validate_bookstore(settings) == expected


def test_validate_bookstore_workspace():
    """Tests that bookstore does not validate with an empty workspace_prefix."""
    expected = {
        "bookstore_valid": True,
        "publish_valid": True,
        "archive_valid": False,
        "clone_valid": True,
    }
    settings = BookstoreSettings(s3_bucket="A_bucket", workspace_prefix="")
    assert validate_bookstore(settings) == expected


def test_validate_bookstore_endpoint():
    """Tests that bookstore does not validate with an empty s3_endpoint_url."""
    expected = {
        "bookstore_valid": False,
        "publish_valid": False,
        "archive_valid": False,
        "clone_valid": True,
    }
    settings = BookstoreSettings(s3_endpoint_url="")
    assert validate_bookstore(settings) == expected


def test_validate_bookstore_bucket():
    """Tests that all bookstore features validate with an s3_bucket."""
    expected = {
        "bookstore_valid": True,
        "publish_valid": True,
        "archive_valid": True,
        "clone_valid": True,
    }
    settings = BookstoreSettings(s3_bucket="A_bucket")
    assert validate_bookstore(settings) == expected


def test_disable_cloning():
    """Tests that all bookstore features validate with an s3_bucket."""
    expected = {
        "bookstore_valid": True,
        "publish_valid": True,
        "archive_valid": True,
        "clone_valid": False,
    }
    settings = BookstoreSettings(s3_bucket="A_bucket", enable_cloning=False)
    assert validate_bookstore(settings) == expected



"""S3 path utilities"""


# Our S3 path delimiter will remain fixed as '/' in all uses
delimiter = "/"


def _join(*args):
    """Join S3 bucket args together.

    Remove empty entries and strip left-leading ``/``
    """
    return delimiter.join(filter(lambda s: s != '', map(lambda s: s.lstrip(delimiter), args)))


def s3_path(bucket, prefix, path=''):
    """Compute the s3 path.

    Parameters
    ----------
    bucket : str
      S3 bucket name
    prefix : str
      prefix for workspace or publish
    path : str
      The storage location
    """
    return _join(bucket, prefix, path)


def s3_key(prefix, path=''):
    """Compute the s3 key

    Parameters
    ----------
    prefix : str
      prefix for workspace or publish
    path : str
      The storage location
    """
    return _join(prefix, path)


def s3_display_path(bucket, prefix, path=''):
    """Create a display name for use in logs

    Parameters
    ----------
    bucket : str
      S3 bucket name
    prefix : str
      prefix for workspace or publish
    path : str
      The storage location
    """
    return 's3://' + s3_path(bucket, prefix, path)



import pytest

from bookstore.client.nb_client import NotebookSession, KernelInfo, LiveNotebookRecord


@pytest.fixture
def notebook_server_dict():
    notebook_server_dict = {
        "base_url": "/",
        "hostname": "localhost",
        "notebook_dir": "/Users/username",
        "password": False,
        "pid": 20981,
        "port": 8888,
        "secure": False,
        "token": "e5814788aeef225172364fcdf1240b90729169a2ced375c7",
        "url": "http://localhost:8888/",
    }
    return notebook_server_dict


@pytest.fixture
def notebook_server_record(notebook_server_dict):
    notebook_server_record = LiveNotebookRecord(**notebook_server_dict)
    return notebook_server_record


@pytest.fixture(scope="module")
def kernel_info_dict():
    info_dict = {
        "id": 'f92b7c8b-0858-4d10-903c-b0631540fb36',
        "name": 'dev',
        "last_activity": '2019-03-14T23:38:08.137987Z',
        "execution_state": 'idle',
        "connections": 0,
    }
    return info_dict


@pytest.fixture(scope="module")
def kernel_info(kernel_info_dict):
    kernel_info = KernelInfo(**kernel_info_dict)
    return kernel_info


@pytest.fixture
def session_dict(kernel_info_dict):
    session_dict = {
        "id": '68d9c58f-c57d-4133-8b41-5ec2731b268d',
        "path": 'Untitled38.ipynb',
        "name": '',
        "type": 'notebook',
        "kernel": kernel_info_dict,
        "notebook": {'path': 'Untitled38.ipynb', 'name': ''},  # deprecated API
    }
    return session_dict


@pytest.fixture
def notebook_session(session_dict):
    notebook_session = NotebookSession(**session_dict)
    return notebook_session



"""Handlers for Bookstore API"""
import json

from notebook.base.handlers import APIHandler
from notebook.base.handlers import path_regex
from .utils import url_path_join
from tornado import web

from ._version import __version__
from .bookstore_config import BookstoreSettings
from .bookstore_config import validate_bookstore
from .publish import BookstorePublishAPIHandler
from .clone import BookstoreCloneHandler, BookstoreCloneAPIHandler


version = __version__


class BookstoreVersionHandler(APIHandler):
    """Handler responsible for Bookstore version information

    Used to lay foundations for the bookstore package. Though, frontends can use this endpoint for feature detection.

    Methods
    -------
    get(self)
        Provides version info and feature availability based on serverside settings.
    build_response_dict(self)
        Helper to populate response.
    """

    @web.authenticated
    def get(self):
        """GET /api/bookstore/

        Returns version info and validation info for various bookstore features.
        """
        self.finish(json.dumps(self.build_response_dict()))

    def build_response_dict(self):
        """Helper for building the version handler's response before serialization."""
        return {
            "release": self.settings['bookstore']["release"],
            "features": self.settings['bookstore']["features"],
        }


def build_settings_dict(validation):
    """Helper for building the settings info that will be assigned to the web_app."""
    return {"release": version, "features": validation}


def load_jupyter_server_extension(nb_app):
    web_app = nb_app.web_app
    host_pattern = '.*$'

    base_url = web_app.settings['base_url']

    bookstore_settings = BookstoreSettings(parent=nb_app)
    validation = validate_bookstore(bookstore_settings)
    web_app.settings['bookstore'] = build_settings_dict(validation)
    handlers = collect_handlers(nb_app.log, base_url, validation)
    web_app.add_handlers(host_pattern, handlers)


def collect_handlers(log, base_url, validation):
    """Utility that collects bookstore endpoints & handlers to be added to the webapp.

    This uses bookstore feature validation to determine which endpoints should be enabled.
    It returns all valid pairs of endpoint patterns and handler classes. 

    Parameters
    ----------
    log : logging.Logger
      Log (usually from the NotebookApp) for logging endpoint changes.
    base_url: str
      The base_url to which we append routes.
    validation: dict
       Validation dictionary for determining which endpoints to enable.

    Returns
    --------
    
    List[Tuple[str, tornado.web.RequestHandler]]
      List of pairs of endpoint patterns and the handler used to handle requests at that endpoint.
    """
    base_bookstore_pattern = url_path_join(base_url, '/bookstore')
    base_bookstore_api_pattern = url_path_join(base_url, '/api/bookstore')

    handlers = []
    # Always enable the version handler for the API
    handlers.append((base_bookstore_api_pattern, BookstoreVersionHandler))

    if validation['publish_valid']:
        log.info(f"[bookstore] Enabling bookstore publishing, version: {version}")
        handlers.append(
            (
                url_path_join(base_bookstore_api_pattern, r"/publish%s" % path_regex),
                BookstorePublishAPIHandler,
            )
        )
    else:
        log.info("[bookstore] Publishing disabled. s3_bucket or endpoint are not configured.")

    if validation['clone_valid']:
        log.info(f"[bookstore] Enabling bookstore cloning, version: {version}")
        handlers.append(
            (url_path_join(base_bookstore_api_pattern, r"/clone(?:/?)*"), BookstoreCloneAPIHandler)
        ),
        handlers.append(
            (url_path_join(base_bookstore_pattern, r"/clone(?:/?)*"), BookstoreCloneHandler)
        )
    else:
        log.info(f"[bookstore] bookstore cloning disabled, version: {version}")

    return handlers



"""Test version checking"""
import logging

import pytest

from .._version import _check_version


@pytest.mark.parametrize(
    'bookstore_version, msg',
    [
        ('', 'Bookstore has no version header'),
        ('1.0.0', 'deprecated bookstore'),
        ('2.0.0', 'Bookstore version is'),
        ('2.3.1', 'Bookstore version is'),
        ('xxxxx', 'Invalid version'),
    ],
)
def test_check_version(bookstore_version, msg, caplog):
    log = logging.getLogger()
    caplog.set_level(logging.DEBUG)
    _check_version(bookstore_version, log)
    record = caplog.records[0]
    assert msg in record.getMessage()



import os

PACKAGE_DIR: str = os.path.realpath(os.path.dirname(__file__))

del os

from .archive import BookstoreContentsArchiver
from .bookstore_config import BookstoreSettings
from .handlers import load_jupyter_server_extension
from ._version import __version__
from ._version import version_info


def _jupyter_server_extension_paths():
    return [dict(module="bookstore")]



"""Tests for s3 paths"""
import pytest

from bookstore.s3_paths import s3_display_path, s3_key


def test_s3_paths():
    bucket = 'mybucket'
    prefix = 'yo'
    path = 'pickles'
    assert s3_display_path(bucket, prefix, path) == 's3://mybucket/yo/pickles'


def test_s3_paths_no_path():
    bucket = 'mybucket'
    prefix = 'yo'
    assert s3_display_path(bucket, prefix) == 's3://mybucket/yo'


def test_s3_paths_no_prefix():
    bucket = 'mybucket'
    path = 'pickles'
    with pytest.raises(NameError):
        s3_display_path(bucket, prefix, path)


def test_s3_key_no_prefix():
    with pytest.raises(TypeError):
        s3_key(path='s3://mybucket')


def test_s3_key_invalid_prefix():
    prefix = 1234
    with pytest.raises(AttributeError):
        s3_key(prefix)


def test_s3_key_valid_parameters():
    prefix = 'workspace'
    path = 'project/manufacturing'
    assert s3_key(prefix, path) == 'workspace/project/manufacturing'



"""Tests for handlers"""
from unittest.mock import Mock

import pytest
import logging
from unittest.mock import Mock

from bookstore._version import __version__
from bookstore.handlers import collect_handlers, build_settings_dict, BookstoreVersionHandler
from bookstore.bookstore_config import BookstoreSettings, validate_bookstore
from bookstore.clone import BookstoreCloneHandler, BookstoreCloneAPIHandler
from bookstore.publish import BookstorePublishAPIHandler
from notebook.base.handlers import path_regex
from tornado.testing import AsyncTestCase
from tornado.web import Application, HTTPError
from tornado.httpserver import HTTPRequest
from traitlets.config import Config

log = logging.getLogger('test_handlers')
version = __version__

from traitlets.config import Config


def test_collect_handlers_all():
    expected = [
        ('/api/bookstore', BookstoreVersionHandler),
        ('/api/bookstore/publish%s' % path_regex, BookstorePublishAPIHandler),
        ('/api/bookstore/clone(?:/?)*', BookstoreCloneAPIHandler),
        ('/bookstore/clone(?:/?)*', BookstoreCloneHandler),
    ]
    web_app = Application()
    mock_settings = {"BookstoreSettings": {"s3_bucket": "mock_bucket"}}
    bookstore_settings = BookstoreSettings(config=Config(mock_settings))
    validation = validate_bookstore(bookstore_settings)
    handlers = collect_handlers(log, '/', validation)
    assert expected == handlers


def test_collect_handlers_no_clone():
    expected = [
        ('/api/bookstore', BookstoreVersionHandler),
        ('/api/bookstore/publish%s' % path_regex, BookstorePublishAPIHandler),
    ]
    web_app = Application()
    mock_settings = {"BookstoreSettings": {"s3_bucket": "mock_bucket", "enable_cloning": False}}
    bookstore_settings = BookstoreSettings(config=Config(mock_settings))
    validation = validate_bookstore(bookstore_settings)
    handlers = collect_handlers(log, '/', validation)
    assert expected == handlers


def test_collect_handlers_no_publish():
    expected = [
        ('/api/bookstore', BookstoreVersionHandler),
        ('/api/bookstore/clone(?:/?)*', BookstoreCloneAPIHandler),
        ('/bookstore/clone(?:/?)*', BookstoreCloneHandler),
    ]
    web_app = Application()
    mock_settings = {"BookstoreSettings": {"s3_bucket": "mock_bucket", "published_prefix": ""}}
    bookstore_settings = BookstoreSettings(config=Config(mock_settings))
    validation = validate_bookstore(bookstore_settings)
    handlers = collect_handlers(log, '/', validation)
    assert expected == handlers


def test_collect_handlers_only_version():
    expected = [('/api/bookstore', BookstoreVersionHandler)]
    web_app = Application()
    mock_settings = {"BookstoreSettings": {"enable_cloning": False}}
    bookstore_settings = BookstoreSettings(config=Config(mock_settings))
    validation = validate_bookstore(bookstore_settings)
    handlers = collect_handlers(log, '/', validation)
    assert expected == handlers


@pytest.fixture(scope="class")
def bookstore_settings(request):
    mock_settings = {
        "BookstoreSettings": {
            "s3_access_key_id": "mock_id",
            "s3_secret_access_key": "mock_access",
            "s3_bucket": "my_bucket",
        }
    }
    config = Config(mock_settings)
    bookstore_settings = BookstoreSettings(config=config)
    if request.cls is not None:
        request.cls.bookstore_settings = bookstore_settings
    return bookstore_settings


def test_build_settings_dict(bookstore_settings):
    expected = {
        'features': {
            'archive_valid': True,
            'bookstore_valid': True,
            'publish_valid': True,
            'clone_valid': True,
        },
        'release': version,
    }
    validation = validate_bookstore(bookstore_settings)
    assert expected == build_settings_dict(validation)


@pytest.mark.usefixtures("bookstore_settings")
class TestCloneAPIHandler(AsyncTestCase):
    def setUp(self):
        super().setUp()

        validation = validate_bookstore(self.bookstore_settings)
        self.mock_application = Mock(
            spec=Application,
            ui_methods={},
            ui_modules={},
            settings={"bookstore": build_settings_dict(validation)},
            transforms=[],
        )

    def get_handler(self, uri, app=None):
        if app is None:
            app = self.mock_application
        connection = Mock(context=Mock(protocol="https"))
        payload_request = HTTPRequest(
            method='GET',
            uri=uri,
            headers={"Host": "localhost:8888"},
            body=None,
            connection=connection,
        )
        return BookstoreVersionHandler(app, payload_request)

    def test_get(self):
        """This is a simple test of the get API at /api/bookstore

        The most notable feature is the need to set _transforms on the handler.

        The default value of handler()._transforms is `None`.
        This is iterated over when handler().flush() is called, raising a TypeError.
        
        In normal usage, the application assigns this when it creates a handler delegate.

        Because our mock application does not do this 
        As a result this raises an error when self.finish() (and therefore self.flush()) is called.

        At runtime on a live Jupyter server, application.transforms == [].
        """
        get_handler = self.get_handler('/api/bookstore/')
        setattr(get_handler, '_transforms', [])
        return_val = get_handler.get()
        assert return_val is None

    def test_build_response(self):
        empty_handler = self.get_handler('/api/bookstore/')
        expected = {
            'features': {
                'archive_valid': True,
                'bookstore_valid': True,
                'publish_valid': True,
                'clone_valid': True,
            },
            'release': version,
        }
        assert empty_handler.build_response_dict() == expected



"""Bookstore version info

Use pep 440 version rules.

No dot before alpha/beta/rc. Use dot before `.dev`. Examples::

- 0.1.0rc1
- 0.1.0a1
- 0.1.0b1.dev
- 0.1.0.dev

`version_info` tuple::

- major
- minor
- micro
- type of release (b1, rc1, or "" for final or dev)
- suffix (dev or "" to designate a final version)
"""

version_info = (2, 3, 0)

__version__ = ".".join(map(str, version_info[:3])) + ".".join(version_info[3:])


def _check_version(bookstore_version, log):
    """Check version and log status"""
    if not bookstore_version:
        log.warning(
            f"Bookstore has no version header, which means it is likely < 2.0. Expected {__version__}"
        )
    elif bookstore_version[:1].isdigit() is False:
        log.warning(f"Invalid version format. Expected {__version__}")
    else:
        if bookstore_version[:1] < '2':
            log.warning(
                f"{bookstore_version} is the deprecated bookstore project for Openstack. Expected {__version__}"
            )
        else:
            log.debug(f"Bookstore version is {bookstore_version}.")
